#
# Holds settings that can be checked
#

# Set to True if running in code coverage mode (only set when running tests)
# -> Since pyfrc 2014.6.0
coverage_mode = False

# Indicates how pyfrc was run: netsim/sim/upload/wipe, etc...
# -> Since pyfrc 2014.6.0
mode = None

# A dictionary of configuration information that is partially loaded
# from sim/config.json
# -> Since pyfrc 2018.1.0
config_obj = {}
